/**CFile***********************************************************************

  FileName    [compileUtil.c]

  PackageName [compile]

  Synopsis    [Routines for model computation.]

  Description [This file contains the code for the compilation of the
  flattened hierarchy into BDD:
  <ul>
  <li> Creation of the boolean variables.</li>
  <li> Creation of the BDD representing the inertia of the system when
       there are processes. In fact when a process is running the
       other processes are stopped, and their state variables don't
       change.</li>
  <li> Creation of the BDD representing what does not change in the
       system, i.e. the set of invariance. These are introduced in the
       model by the keyword "<tt>INVAR</tt>" or by the <em>normal
       assignments</em> (i.e. "<tt>ASSIGN x : = y & z;</tt>"). These
       states are not stored in the transition relation, they are
       stored in an a doc variable.
  <li> Creation of the BDD representing the set of initial states.
  <li> Creation of the BDD representing the transition relation. 
       Various ways of representing the transition relation are offered
       the users.
       <ul>
       <li> <em>Monolithic</em>: the monolithic transition relation is
            computed.</li>
       <li> <em>Conjunctive Partioned (Threshold)</em>: the transition 
            relation is stored as an implicitly conjoined list of 
            transition relation. This kind of partitioning can be 
            used only if the model considered is a synchronous one.</li>
       <li> <em>Conjunctive Partioned IWLS95</em>: as the above, but the
            heuristic proposed in \[1\] is used to order partition clusters. </li>
       </ul>
  <li> Computes the fairness constraints. I.e. each fairness constraint
       (which can be a CTL formula) is evaluated and the resulting BDD
       is stored in the list <tt>fairness_constraints_bdd</tt> to be
       then used in the model checking phase.
  </ul>
  \[1\] R. K. Ranjan and A. Aziz and B. Plessier and C. Pixley and R. K. Brayton,
      "Efficient BDD Algorithms for FSM Synthesis and Verification,
      IEEE/ACM Proceedings International Workshop on Logic Synthesis,
      Lake Tahoe (NV), May 1995.</li>
  ]

  SeeAlso     []

  Author      [Marco Roveri]

  Copyright   [
  This file is part of the ``compile'' package of NuSMV version 2. 
  Copyright (C) 1998-2001 by CMU and ITC-irst. 

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/
#include "compile/compile.h"
#include "compileInt.h" 
#include "parser/symbols.h"
#include "utils/ustring.h"
#include "utils/error.h"


static char rcsid[] UTIL_UNUSED = "$Id: compileUtil.c,v 1.29.4.12.4.8 2006/07/14 07:57:11 nusmv Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

typedef enum {
  State_Instantiation_Mode,
  Input_Instantiation_Mode
} Instantiation_Vars_Mode_Type;



/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/
 
/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Builds an internal representation for a given string.]

  Description        [Builds an internal representation for a given
  string. If the conversion has been performed in the past, then the
  hashed value is returned back, else a new one is created, hashed and
  returned. We hash this in order to allow the following:
  <pre>
  VAR
     x : {a1, a2, a3};
     y : {a3, a4, a5};

  ASSIGN
     next(x) := case
                 x = y    : a2;
                 !(x = y) : a1;
                 1        : a3;
                esac;
  </pre>
  i.e. to allow the equality test between x and y. This can be
  performed because we internally have a unique representation of the
  atom <tt>a3</tt>.]

  SideEffects        []

  SeeAlso            [find_atom]

******************************************************************************/
node_ptr sym_intern(char *s)
{
  return find_node(ATOM, (node_ptr)find_string(s), Nil);
}



/**Macro***********************************************************************
  Synopsis           [Private macros for the sake of readability of the function
  Compile_pop_global]

  Description        []
******************************************************************************/
#define IS_AND_DISTRIB_OP(prop)             \
   ((OP_GLOBAL == node_get_type(prop)) ||   \
    (AG == node_get_type(prop)) ||          \
    (OP_HISTORICAL == node_get_type(prop)))

#define IS_OR_DISTRIB_OP(prop)              \
   ((OP_FUTURE == node_get_type(prop)) ||   \
    (AF == node_get_type(prop)) ||          \
    (OP_ONCE == node_get_type(prop)))


#define ARE_AND_DISTRIB_OPS(prop1, prop2)         \
   (IS_AND_DISTRIB_OP(prop1) &&                   \
    IS_AND_DISTRIB_OP(prop2) &&                   \
    node_get_type(prop1) == node_get_type(prop2))

#define ARE_OR_DISTRIB_OPS(prop1, prop2)         \
   (IS_OR_DISTRIB_OP(prop1) &&                   \
    IS_OR_DISTRIB_OP(prop2) &&                   \
    node_get_type(prop1) == node_get_type(prop2))



/**Function********************************************************************

  Synopsis           [Simplifies the given property by exploiting 
  the distributivity of G, AG and H over AND, and distributivity of F, AF and O 
  over OR]

  Description        [Transformation rules are:
   1) <OP> <OP> a           :-> <OP> a
   2) (<OP> a) * (<OP> b)   :-> <OP> (a * b);
   3) (<OP> (a * <OP> b))   :-> <OP> (a * b);
   4) (<OP> (<OP> a * b))   :-> <OP> (a * b);
   5) (<OP> (<OP> a * <OP> b)) :-> <OP> (a * b); 

   Where <OP> can be either:
     G|AG|H for * := &
     F|AF|O for * := |

   Given property can be both flattened or unflattened.
   ]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
node_ptr Compile_pop_distrib_ops(node_ptr prop) 
{
  if ((node_ptr) NULL == prop) return (node_ptr) NULL;

  /* Base cases */
  switch (node_get_type(prop)) {
  case FAILURE:
  case TRUEEXP: 
  case FALSEEXP:
  case SELF: 
  case BOOLEAN: 
  case ATOM:
  case DOT:
  case ARRAY:
  case NUMBER:
  case NUMBER_WORD:
  case NUMBER_FRAC:
  case NUMBER_REAL:
  case NUMBER_EXP:
    return prop;

  case CONTEXT:
    return find_node(CONTEXT, car(prop), Compile_pop_distrib_ops(cdr(prop)));

  default: break;
  }

  /* 1) <OP> <OP> a :-> <OP> a */
  if (ARE_AND_DISTRIB_OPS(prop, car(prop)) || 
      ARE_OR_DISTRIB_OPS(prop, car(prop))) {
    return Compile_pop_distrib_ops(car(prop));
  }

  /* 2) (<OP> a) * (<OP> b) :-> <OP> (a * b); */
  if (AND == node_get_type(prop) || 
      OR == node_get_type(prop)) {
    node_ptr l = Compile_pop_distrib_ops(car(prop));
    node_ptr r = Compile_pop_distrib_ops(cdr(prop));

    if ((ARE_AND_DISTRIB_OPS(l, r) && AND == node_get_type(prop)) ||
	(ARE_OR_DISTRIB_OPS(l, r) && OR == node_get_type(prop))) {
      return Compile_pop_distrib_ops(find_node(node_get_type(l),
		  find_node(node_get_type(prop), car(l), car(r)), Nil));
    }
    else return find_node(node_get_type(prop), l, r);
  }

  /* 3) (<OP> (a * <OP> b))   :-> <OP> (a * b);
     4) (<OP> (<OP> a * b))   :-> <OP> (a * b);
     5) (<OP> (<OP> a * <OP> b)) :-> <OP> (a * b); */

  if (IS_AND_DISTRIB_OP(prop) || IS_OR_DISTRIB_OP(prop)) {
    int op = node_get_type(car(prop));
    if (AND == op || OR == op) {
      node_ptr l = Compile_pop_distrib_ops(car(car(prop)));
      node_ptr r = Compile_pop_distrib_ops(cdr(car(prop)));

      if ( ((AND == op) && ARE_AND_DISTRIB_OPS(prop, l) && 
	    ARE_AND_DISTRIB_OPS(prop, r)) 
	   ||
	   ((OR == op) && ARE_OR_DISTRIB_OPS(prop, l) && 
	    ARE_OR_DISTRIB_OPS(prop, r)) ) { /* 5 */

        return Compile_pop_distrib_ops(find_node(node_get_type(prop),
					 find_node(op, car(l), car(r)), Nil));
      }
      else if ( ((AND == op) && ARE_AND_DISTRIB_OPS(prop, l)) 
		|| 
		((OR == op) && ARE_OR_DISTRIB_OPS(prop, l)) ) { /* 4 */
        return Compile_pop_distrib_ops(find_node(node_get_type(prop), 
					    find_node(op, car(l), r), Nil));
      }
      else if ( ((AND == op) && ARE_AND_DISTRIB_OPS(prop, r)) 
		|| 
		((OR == op) && ARE_OR_DISTRIB_OPS(prop, r)) ) { /* 3 */
        return Compile_pop_distrib_ops(find_node(node_get_type(prop),
					    find_node(op, l, car(r)), Nil));
      }
      return find_node(node_get_type(prop), find_node(op, l, r), Nil);
    }
  }

  /* fall back case */
  return find_node(node_get_type(prop), 
		   Compile_pop_distrib_ops(car(prop)), 
		   Compile_pop_distrib_ops(cdr(prop)));
}


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

